/*
 * Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file command.c Client authentication and server commands. */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <signal.h>
#include <sys/types.h>
#include <osl.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>

#include "para.h"
#include "error.h"
#include "crypt.h"
#include "sideband.h"
#include "command.h"
#include "server.cmdline.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "list.h"
#include "send.h"
#include "sched.h"
#include "vss.h"
#include "net.h"
#include "daemon.h"
#include "fd.h"
#include "ipc.h"
#include "user_list.h"
#include "server.command_list.h"
#include "afs.command_list.h"
#include "signal.h"
#include "version.h"

typedef int server_command_handler_t(struct command_context *);
static server_command_handler_t SERVER_COMMAND_HANDLERS;
server_command_handler_t AFS_COMMAND_HANDLERS;

/* Defines one command of para_server. */
struct server_command {
	/* The name of the command. */
	const char *name;
	/* Pointer to the function that handles the command. */
	server_command_handler_t *handler;
	/* The privileges a user must have to execute this command. */
	unsigned int perms;
	/* One-line description of the command. */
	const char *description;
	/* Summary of the command line options. */
	const char *usage;
	/* The long help text. */
	const char *help;
};

static struct server_command afs_cmds[] = {DEFINE_AFS_CMD_ARRAY};
static struct server_command server_cmds[] = {DEFINE_SERVER_CMD_ARRAY};

/** Commands including options must be shorter than this. */
#define MAX_COMMAND_LEN 32768

extern int mmd_mutex;
extern struct misc_meta_data *mmd;
extern struct sender senders[];
int send_afs_status(struct command_context *cc, int parser_friendly);

static void dummy(__a_unused int s)
{
}

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mutex_lock(mmd_mutex);
	*new_mmd = *mmd;
	mutex_unlock(mmd_mutex);
}

/*
 * Compute human readable string containing vss status for given integer value.
 *
 * We don't want to use vss_playing() and friends here because we take a
 * snapshot of the mmd struct and use the copy for computing the state of the
 * vss. If the real data were used, we would take the mmd lock for a rather
 * long time or risk to get an inconsistent view.
 */
static char *vss_status_tohuman(unsigned int flags)
{
	if (flags & VSS_PLAYING)
		return para_strdup("playing");
	if (flags & VSS_NEXT)
		return para_strdup("stopped");
	return para_strdup("paused");
}

/*
 * return human readable permission string. Never returns NULL.
 */
static char *cmd_perms_itohuman(unsigned int perms)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = perms & AFS_READ? 'a' : '-';
	msg[1] = perms & AFS_WRITE? 'A' : '-';
	msg[2] = perms & VSS_READ? 'v' : '-';
	msg[3] = perms & VSS_WRITE? 'V' : '-';
	msg[4] = '\0';
	return msg;
}

/*
 * Never returns NULL.
 */
static char *vss_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = (flags & VSS_PLAYING)? 'P' : '_';
	msg[1] = (flags & VSS_NOMORE)? 'O' : '_';
	msg[2] = (flags & VSS_NEXT)? 'N' : '_';
	msg[3] = (flags & VSS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}

static unsigned get_status(struct misc_meta_data *nmmd, int parser_friendly,
		char **result)
{
	char *status, *flags; /* vss status info */
	/* nobody updates our version of "now" */
	long offset = (nmmd->offset + 500) / 1000;
	struct timeval current_time;
	struct para_buffer b = {.flags = parser_friendly? PBF_SIZE_PREFIX : 0};

	/* report real status */
	status = vss_status_tohuman(nmmd->vss_status_flags);
	flags = vss_get_status_flags(nmmd->vss_status_flags);
	clock_get_realtime(&current_time);
	/*
	 * The calls to WRITE_STATUS_ITEM() below never fail because
	 * b->max_size is zero (unlimited), see para_printf(). However, clang
	 * is not smart enough to prove this and complains nevertheless.
	 * Casting the return value to void silences clang.
	 */
	(void)WRITE_STATUS_ITEM(&b, SI_STATUS, "%s\n", status);
	(void)WRITE_STATUS_ITEM(&b, SI_STATUS_FLAGS, "%s\n", flags);
	(void)WRITE_STATUS_ITEM(&b, SI_OFFSET, "%li\n", offset);
	(void)WRITE_STATUS_ITEM(&b, SI_AFS_MODE, "%s\n", mmd->afs_mode_string);
	(void)WRITE_STATUS_ITEM(&b, SI_STREAM_START, "%lu.%lu\n",
		(long unsigned)nmmd->stream_start.tv_sec,
		(long unsigned)nmmd->stream_start.tv_usec);
	(void)WRITE_STATUS_ITEM(&b, SI_CURRENT_TIME, "%lu.%lu\n",
		(long unsigned)current_time.tv_sec,
		(long unsigned)current_time.tv_usec);
	free(flags);
	free(status);
	*result = b.buf;
	return b.offset;
}

static int check_sender_args(int argc, char * const * argv, struct sender_command_data *scd)
{
	int i;
	/* this has to match sender.h */
	const char *subcmds[] = {"add", "delete", "allow", "deny", "on", "off", NULL};

	scd->sender_num = -1;
	if (argc < 3)
		return -E_COMMAND_SYNTAX;
	for (i = 0; senders[i].name; i++)
		if (!strcmp(senders[i].name, argv[1]))
			break;
	PARA_DEBUG_LOG("%d:%s\n", argc, argv[1]);
	if (!senders[i].name)
		return -E_COMMAND_SYNTAX;
	scd->sender_num = i;
	for (i = 0; subcmds[i]; i++)
		if (!strcmp(subcmds[i], argv[2]))
			break;
	if (!subcmds[i])
		return -E_COMMAND_SYNTAX;
	scd->cmd_num = i;
	if (!senders[scd->sender_num].client_cmds[scd->cmd_num])
		return -E_SENDER_CMD;
	switch (scd->cmd_num) {
	case SENDER_ON:
	case SENDER_OFF:
		if (argc != 3)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_DENY:
	case SENDER_ALLOW:
		if (argc != 4 || parse_cidr(argv[3], scd->host,
				sizeof(scd->host), &scd->netmask) == NULL)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_ADD:
	case SENDER_DELETE:
		if (argc != 4)
			return -E_COMMAND_SYNTAX;
		return parse_fec_url(argv[3], scd);
	default:
		return -E_COMMAND_SYNTAX;
	}
	return 1;
}

/**
 * Send a sideband packet through a blocking file descriptor.
 *
 * \param scc fd and crypto keys.
 * \param buf The buffer to send.
 * \param numbytes The size of \a buf.
 * \param band The sideband designator of this packet.
 * \param dont_free If true, never deallocate \a buf.
 *
 * The nonblock flag must be disabled for the file descriptor given by \a scc.
 *
 * Stream cipher encryption is automatically activated if necessary via the
 * sideband transformation, depending on the value of \a band.
 *
 * \return Standard.
 *
 * \sa \ref send_sb_va().
 */
int send_sb(struct stream_cipher_context *scc, void *buf, size_t numbytes,
		int band, bool dont_free)
{
	int ret;
	struct sb_context *sbc;
	struct iovec iov[2];
	sb_transformation trafo = band < SBD_PROCEED? NULL : sc_trafo;
	struct sb_buffer sbb = SBB_INIT(band, buf, numbytes);

	sbc = sb_new_send(&sbb, dont_free, trafo, scc->send);
	do {
		ret = sb_get_send_buffers(sbc, iov);
		ret = xwritev(scc->fd, iov, ret);
		if (ret < 0)
			goto fail;
	} while (sb_sent(sbc, ret) == false);
	return 1;
fail:
	sb_free(sbc);
	return ret;
}

/**
 * Create a variable sized buffer and send it as a sideband packet.
 *
 * \param scc Passed to \ref send_sb.
 * \param band See \ref send_sb.
 * \param fmt The format string.
 *
 * \return The return value of the underlying call to \ref send_sb.
 */
__printf_3_4 int send_sb_va(struct stream_cipher_context *scc, int band,
		const char *fmt, ...)
{
	va_list ap;
	char *msg;
	int ret;

	va_start(ap, fmt);
	ret = xvasprintf(&msg, fmt, ap);
	va_end(ap);
	return send_sb(scc, msg, ret, band, false);
}

/**
 * Send an error message to a client.
 *
 * \param cc Client info.
 * \param err The (positive) error code.
 *
 * \return The return value of the underlying call to send_sb_va().
 */
int send_strerror(struct command_context *cc, int err)
{
	return send_sb_va(&cc->scc, SBD_ERROR_LOG, "%s\n", para_strerror(err));
}

/**
 * Send a sideband packet through a blocking file descriptor.
 *
 * \param scc fd and crypto keys.
 * \param expected_band The expected band designator.
 * \param max_size Passed to \ref sb_new_recv().
 * \param result Body of the sideband packet is returned here.
 *
 * If \a expected_band is not \p SBD_ANY, the band designator of the received
 * sideband packet is compared to \a expected_band and a mismatch is considered
 * an error.
 *
 * \return Standard.
 */
int recv_sb(struct stream_cipher_context *scc,
		enum sb_designator expected_band,
		size_t max_size, struct iovec *result)
{
	int ret;
	struct sb_context *sbc;
	struct iovec iov;
	struct sb_buffer sbb;
	sb_transformation trafo;

	trafo = expected_band != SBD_ANY && expected_band < SBD_PROCEED?
		NULL : sc_trafo;
	sbc = sb_new_recv(max_size, trafo, scc->recv);
	for (;;) {
		sb_get_recv_buffer(sbc, &iov);
		ret = recv_bin_buffer(scc->fd, iov.iov_base, iov.iov_len);
		if (ret == 0)
			ret = -E_EOF;
		if (ret < 0)
			goto fail;
		ret = sb_received(sbc, ret, &sbb);
		if (ret < 0)
			goto fail;
		if (ret > 0)
			break;
	}
	ret = -E_BAD_BAND;
	if (expected_band != SBD_ANY && sbb.band != expected_band)
		goto fail;
	*result = sbb.iov;
	return 1;
fail:
	sb_free(sbc);
	return ret;
}

static int com_sender(struct command_context *cc)
{
	int i, ret = 0;
	char *msg = NULL;
	struct sender_command_data scd;

	if (cc->argc < 2) {
		for (i = 0; senders[i].name; i++) {
			char *tmp;
			ret = xasprintf(&tmp, "%s%s\n", msg? msg : "",
				senders[i].name);
			free(msg);
			msg = tmp;
		}
		return send_sb(&cc->scc, msg, ret, SBD_OUTPUT, false);
	}
	ret = check_sender_args(cc->argc, cc->argv, &scd);
	if (ret < 0) {
		if (scd.sender_num < 0)
			return ret;
		if (strcmp(cc->argv[2], "status") == 0)
			msg = senders[scd.sender_num].status();
		else
			msg = senders[scd.sender_num].help();
		return send_sb(&cc->scc, msg, strlen(msg), SBD_OUTPUT, false);
	}

	switch (scd.cmd_num) {
	case SENDER_ADD:
	case SENDER_DELETE:
		assert(senders[scd.sender_num].resolve_target);
		ret = senders[scd.sender_num].resolve_target(cc->argv[3], &scd);
		if (ret < 0)
			return ret;
	}

	for (i = 0; i < 10; i++) {
		mutex_lock(mmd_mutex);
		if (mmd->sender_cmd_data.cmd_num >= 0) {
			/* another sender command is active, retry in 100ms */
			struct timespec ts = {.tv_nsec = 100 * 1000 * 1000};
			mutex_unlock(mmd_mutex);
			nanosleep(&ts, NULL);
			continue;
		}
		mmd->sender_cmd_data = scd;
		mutex_unlock(mmd_mutex);
		break;
	}
	return (i < 10)? 1 : -E_LOCK;
}

/* server info */
static int com_si(struct command_context *cc)
{
	int ret;
	char *msg, *ut;

	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	ut = daemon_get_uptime_str(now);
	ret = xasprintf(&msg,
		"up: %s\nplayed: %u\n"
		"server_pid: %d\n"
		"afs_pid: %d\n"
		"connections (active/accepted/total): %u/%u/%u\n"
		"current loglevel: %s\n"
		"supported audio formats: %s\n",
		ut, mmd->num_played,
		(int)getppid(),
		(int)mmd->afs_pid,
		mmd->active_connections,
		mmd->num_commands,
		mmd->num_connects,
		conf.loglevel_arg,
		AUDIO_FORMAT_HANDLERS
	);
	mutex_unlock(mmd_mutex);
	free(ut);
	return send_sb(&cc->scc, msg, ret, SBD_OUTPUT, false);
}

/* version */
static int com_version(struct command_context *cc)
{
	char *msg;
	size_t len;

	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	len = xasprintf(&msg, "%s", version_text("server"));
	return send_sb(&cc->scc, msg, len, SBD_OUTPUT, false);
}

/** These status items are cleared if no audio file is currently open. */
#define EMPTY_STATUS_ITEMS \
	ITEM(PATH) \
	ITEM(DIRECTORY) \
	ITEM(BASENAME) \
	ITEM(SCORE) \
	ITEM(ATTRIBUTES_BITMAP) \
	ITEM(ATTRIBUTES_TXT) \
	ITEM(HASH) \
	ITEM(IMAGE_ID) \
	ITEM(IMAGE_NAME) \
	ITEM(LYRICS_ID) \
	ITEM(LYRICS_NAME) \
	ITEM(BITRATE) \
	ITEM(FORMAT) \
	ITEM(FREQUENCY) \
	ITEM(CHANNELS) \
	ITEM(DURATION) \
	ITEM(SECONDS_TOTAL) \
	ITEM(NUM_PLAYED) \
	ITEM(LAST_PLAYED) \
	ITEM(TECHINFO) \
	ITEM(ARTIST) \
	ITEM(TITLE) \
	ITEM(YEAR) \
	ITEM(ALBUM) \
	ITEM(COMMENT) \
	ITEM(MTIME) \
	ITEM(FILE_SIZE) \
	ITEM(CHUNK_TIME) \
	ITEM(NUM_CHUNKS) \
	ITEM(AMPLIFICATION) \

/**
 * Write a list of audio-file related status items with empty values.
 *
 * This is used by vss when currently no audio file is open.
 */
static unsigned empty_status_items(int parser_friendly, char **result)
{
	char *esi;
	unsigned len;

	if (parser_friendly)
		len = xasprintf(&esi,
			#define ITEM(x) "0004 %02x:\n"
			EMPTY_STATUS_ITEMS
			#undef ITEM
			#define ITEM(x) , SI_ ## x
			EMPTY_STATUS_ITEMS
			#undef ITEM
		);
	else
		len = xasprintf(&esi,
			#define ITEM(x) "%s:\n"
			EMPTY_STATUS_ITEMS
			#undef ITEM
			#define ITEM(x) ,status_item_list[SI_ ## x]
			EMPTY_STATUS_ITEMS
			#undef ITEM
		);
	*result = esi;
	return len;
}
#undef EMPTY_STATUS_ITEMS

/* stat */
static int com_stat(struct command_context *cc)
{
	int i, ret;
	struct misc_meta_data tmp, *nmmd = &tmp;
	char *s;
	int32_t num = 0;
	int parser_friendly = 0;

	para_sigaction(SIGUSR1, dummy);

	for (i = 1; i < cc->argc; i++) {
		const char *arg = cc->argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strncmp(arg, "-n=", 3)) {
			ret = para_atoi32(arg + 3, &num);
			if (ret < 0)
				return ret;
			continue;
		}
		if (!strcmp(arg, "-p")) {
			parser_friendly = 1;
			continue;
		}
		return -E_COMMAND_SYNTAX;
	}
	if (i != cc->argc)
		return -E_COMMAND_SYNTAX;
	for (;;) {
		mmd_dup(nmmd);
		ret = get_status(nmmd, parser_friendly, &s);
		ret = send_sb(&cc->scc, s, ret, SBD_OUTPUT, false);
		if (ret < 0)
			goto out;
		if (nmmd->vss_status_flags & VSS_NEXT) {
			char *esi;
			ret = empty_status_items(parser_friendly, &esi);
			ret = send_sb(&cc->scc, esi, ret, SBD_OUTPUT, false);
			if (ret < 0)
				goto out;
		} else
			send_afs_status(cc, parser_friendly);
		ret = 1;
		if (num > 0 && !--num)
			goto out;
		sleep(50);
		ret = -E_SERVER_CRASH;
		if (getppid() == 1)
			goto out;
	}
out:
	return ret;
}

static int send_list_of_commands(struct command_context *cc, struct server_command *cmd,
		const char *handler)
{
	char *msg = NULL;

	for (; cmd->name; cmd++) {
		char *tmp, *perms = cmd_perms_itohuman(cmd->perms);
		tmp = make_message("%s\t%s\t%s\t%s\n", cmd->name, handler,
			perms, cmd->description);
		free(perms);
		msg = para_strcat(msg, tmp);
		free(tmp);
	}
	assert(msg);
	return send_sb(&cc->scc, msg, strlen(msg), SBD_OUTPUT, false);
}

/* returns string that must be freed by the caller */
static struct server_command *get_cmd_ptr(const char *name, char **handler)
{
	struct server_command *cmd;

	for (cmd = server_cmds; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("server"); /* server commands */
			return cmd;
		}
	/* not found, look for commands supported by afs */
	for (cmd = afs_cmds; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("afs");
			return cmd;
		}
	return NULL;
}

/* help */
static int com_help(struct command_context *cc)
{
	struct server_command *cmd;
	char *perms, *handler, *buf;
	int ret;

	if (cc->argc < 2) {
		/* no argument given, print list of commands */
		if ((ret = send_list_of_commands(cc, server_cmds, "server")) < 0)
			return ret;
		return send_list_of_commands(cc, afs_cmds, "afs");
	}
	/* argument given for help */
	cmd = get_cmd_ptr(cc->argv[1], &handler);
	if (!cmd)
		return -E_BAD_CMD;
	perms = cmd_perms_itohuman(cmd->perms);
	ret = xasprintf(&buf, "%s - %s\n\n"
		"handler: %s\n"
		"permissions: %s\n"
		"usage: %s\n\n"
		"%s\n",
		cc->argv[1],
		cmd->description,
		handler,
		perms,
		cmd->usage,
		cmd->help
	);
	free(perms);
	free(handler);
	return send_sb(&cc->scc, buf, ret, SBD_OUTPUT, false);
}

/* hup */
static int com_hup(struct command_context *cc)
{
	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGHUP);
	return 1;
}

/* term */
static int com_term(struct command_context *cc)
{
	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGTERM);
	return 1;
}

static int com_play(struct command_context *cc)
{
	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* stop */
static int com_stop(struct command_context *cc)
{
	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags &= ~VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_REPOS;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* pause */
static int com_pause(struct command_context *cc)
{
	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	if (!vss_paused() && !vss_stopped()) {
		mmd->events++;
		mmd->new_vss_status_flags &= ~VSS_PLAYING;
		mmd->new_vss_status_flags &= ~VSS_NEXT;
	}
	mutex_unlock(mmd_mutex);
	return 1;
}

/* next */
static int com_next(struct command_context *cc)
{
	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->events++;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* nomore */
static int com_nomore(struct command_context *cc)
{
	if (cc->argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	if (vss_playing() || vss_paused())
		mmd->new_vss_status_flags |= VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* ff */
static int com_ff(struct command_context *cc)
{
	long promille;
	int ret, backwards = 0;
	unsigned i;
	char c;

	if (cc->argc != 2)
		return -E_COMMAND_SYNTAX;
	if (!(ret = sscanf(cc->argv[1], "%u%c", &i, &c)))
		return -E_COMMAND_SYNTAX;
	if (ret > 1 && c == '-')
		backwards = 1; /* jmp backwards */
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total || !mmd->afd.afhi.seconds_total)
		goto out;
	promille = (1000 * mmd->current_chunk) / mmd->afd.afhi.chunks_total;
	if (backwards)
		promille -= 1000 * i / mmd->afd.afhi.seconds_total;
	else
		promille += 1000 * i / mmd->afd.afhi.seconds_total;
	if (promille < 0)
		promille = 0;
	if (promille > 1000) {
		mmd->new_vss_status_flags |= VSS_NEXT;
		goto out;
	}
	mmd->repos_request = (mmd->afd.afhi.chunks_total * promille) / 1000;
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	mmd->events++;
	ret = 1;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}

/* jmp */
static int com_jmp(struct command_context *cc)
{
	long unsigned int i;
	int ret;

	if (cc->argc != 2)
		return -E_COMMAND_SYNTAX;
	if (sscanf(cc->argv[1], "%lu", &i) <= 0)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total)
		goto out;
	if (i > 100)
		i = 100;
	PARA_INFO_LOG("jumping to %lu%%\n", i);
	mmd->repos_request = (mmd->afd.afhi.chunks_total * i + 50) / 100;
	PARA_INFO_LOG("sent: %lu, offset before jmp: %lu\n",
		mmd->chunks_sent, mmd->offset);
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	ret = 1;
	mmd->events++;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}

static int com_tasks(struct command_context *cc)
{
	char *tl = server_get_tasks();
	int ret = 1;

	if (tl)
		ret = send_sb(&cc->scc, tl, strlen(tl), SBD_OUTPUT, false);
	return ret;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms.
 * Returns 0 if perms are sufficient, -E_PERM otherwise.
 */
static int check_perms(unsigned int perms, const struct server_command *cmd_ptr)
{
	PARA_DEBUG_LOG("checking permissions\n");
	return (cmd_ptr->perms & perms) < cmd_ptr->perms ? -E_PERM : 0;
}

static void reset_signals(void)
{
	para_sigaction(SIGCHLD, SIG_IGN);
	para_sigaction(SIGINT, SIG_DFL);
	para_sigaction(SIGTERM, SIG_DFL);
	para_sigaction(SIGHUP, SIG_DFL);
}

struct connection_features {
	bool sideband_requested;
	bool aes_ctr128_requested;
};

static int parse_auth_request(char *buf, int len, struct user **u,
		struct connection_features *cf)
{
	int ret;
	char *p, *username, **features = NULL;
	size_t auth_rq_len = strlen(AUTH_REQUEST_MSG);

	*u = NULL;
	memset(cf, 0, sizeof(*cf));
	if (len < auth_rq_len + 2)
		return -E_AUTH_REQUEST;
	if (strncmp(buf, AUTH_REQUEST_MSG, auth_rq_len) != 0)
		return -E_AUTH_REQUEST;
	username = buf + auth_rq_len;
	p = strchr(username, ' ');
	if (p) {
		int i;
		if (p == username)
			return -E_AUTH_REQUEST;
		*p = '\0';
		p++;
		create_argv(p, ",", &features);
		for (i = 0; features[i]; i++) {
			if (strcmp(features[i], "sideband") == 0)
				cf->sideband_requested = true;
			else if (strcmp(features[i], "aes_ctr128") == 0)
				cf->aes_ctr128_requested = true;
			else {
				ret = -E_BAD_FEATURE;
				goto out;
			}
		}
	}
	PARA_DEBUG_LOG("received auth request for user %s\n", username);
	*u = lookup_user(username);
	ret = 1;
out:
	free_argv(features);
	return ret;
}

#define HANDSHAKE_BUFSIZE 4096

static int run_command(struct command_context *cc, struct iovec *iov,
		const char *peername)
{
	int ret, i;
	char *p, *end;
	struct server_command *cmd;

	if (iov->iov_base == NULL || iov->iov_len == 0)
		return -E_BAD_CMD;
	p = iov->iov_base;
	p[iov->iov_len - 1] = '\0'; /* just to be sure */
	cmd = get_cmd_ptr(p, NULL);
	if (!cmd)
		return -E_BAD_CMD;
	ret = check_perms(cc->u->perms, cmd);
	if (ret < 0)
		return ret;
	end = iov->iov_base + iov->iov_len;
	for (i = 0; p < end; i++)
		p += strlen(p) + 1;
	cc->argc = i;
	cc->argv = para_malloc((cc->argc + 1) * sizeof(char *));
	for (i = 0, p = iov->iov_base; p < end; i++) {
		cc->argv[i] = para_strdup(p);
		p += strlen(p) + 1;
	}
	cc->argv[cc->argc] = NULL;
	PARA_NOTICE_LOG("calling com_%s() for %s@%s\n", cmd->name,
		cc->u->name, peername);
	ret = cmd->handler(cc);
	free_argv(cc->argv);
	mutex_lock(mmd_mutex);
	mmd->num_commands++;
	if (ret >= 0 && (cmd->perms & AFS_WRITE))
		mmd->events++;
	mutex_unlock(mmd_mutex);
	return ret;
}

/**
 * Perform user authentication and execute a command.
 *
 * \param fd The file descriptor to send output to.
 * \param peername Identifies the connecting peer.
 *
 * Whenever para_server accepts an incoming tcp connection on the port it
 * listens on, it forks and the resulting child calls this function.
 *
 * An RSA-based challenge/response is used to authenticate the peer. It that
 * authentication succeeds, a random session key is generated and sent back to
 * the peer, encrypted with its RSA public key. From this point on, all
 * transfers are crypted with this session key.
 *
 * Next it is checked if the peer supplied a valid server command or a command
 * for the audio file selector. If yes, and if the user has sufficient
 * permissions to execute that command, the function calls the corresponding
 * command handler which does argument checking and further processing.
 *
 * In order to cope with a DOS attacks, a timeout is set up which terminates
 * the function if the connection was not authenticated when the timeout
 * expires.
 *
 * \sa alarm(2), crypt.c, crypt.h
 */
__noreturn void handle_connect(int fd, const char *peername)
{
	int ret;
	unsigned char rand_buf[CHALLENGE_SIZE + 2 * SESSION_KEY_LEN];
	unsigned char challenge_hash[HASH_SIZE];
	char *command = NULL, *buf = para_malloc(HANDSHAKE_BUFSIZE) /* must be on the heap */;
	size_t numbytes;
	struct command_context cc_struct = {.peer = peername}, *cc = &cc_struct;
	struct iovec iov;
	struct connection_features cf;

	cc->scc.fd = fd;
	reset_signals();
	/* we need a blocking fd here as recv() might return EAGAIN otherwise. */
	ret = mark_fd_blocking(fd);
	if (ret < 0)
		goto net_err;
	/* send Welcome message */
	ret = write_va_buffer(fd, "This is para_server, version "
		PACKAGE_VERSION ".\n"
		"Features: sideband,aes_ctr128\n"
	);
	if (ret < 0)
		goto net_err;
	/* recv auth request line */
	ret = recv_buffer(fd, buf, HANDSHAKE_BUFSIZE);
	if (ret < 0)
		goto net_err;
	ret = parse_auth_request(buf, ret, &cc->u, &cf);
	if (ret < 0)
		goto net_err;
	if (!cf.sideband_requested) { /* sideband is mandatory */
		PARA_ERROR_LOG("client did not request sideband\n");
		ret = -E_BAD_FEATURE;
		goto net_err;
	}
	if (cc->u) {
		get_random_bytes_or_die(rand_buf, sizeof(rand_buf));
		ret = pub_encrypt(cc->u->pubkey, rand_buf, sizeof(rand_buf),
			(unsigned char *)buf);
		if (ret < 0)
			goto net_err;
		numbytes = ret;
	} else {
		/*
		 * We don't want to reveal our user names, so we send a
		 * challenge to the client even if the user does not exist, and
		 * fail the authentication later.
		 */
		numbytes = 256;
		get_random_bytes_or_die((unsigned char *)buf, numbytes);
	}
	PARA_DEBUG_LOG("sending %u byte challenge + session key (%zu bytes)\n",
		CHALLENGE_SIZE, numbytes);
	ret = send_sb(&cc->scc, buf, numbytes, SBD_CHALLENGE, false);
	buf = NULL;
	if (ret < 0)
		goto net_err;
	ret = recv_sb(&cc->scc, SBD_CHALLENGE_RESPONSE,
		HANDSHAKE_BUFSIZE, &iov);
	if (ret < 0)
		goto net_err;
	buf = iov.iov_base;
	numbytes = iov.iov_len;
	PARA_DEBUG_LOG("received %zu bytes challenge response\n", numbytes);
	ret = -E_BAD_USER;
	if (!cc->u)
		goto net_err;
	/*
	 * The correct response is the hash of the first CHALLENGE_SIZE bytes
	 * of the random data.
	 */
	ret = -E_BAD_AUTH;
	if (numbytes != HASH_SIZE)
		goto net_err;
	hash_function((char *)rand_buf, CHALLENGE_SIZE, challenge_hash);
	if (memcmp(challenge_hash, buf, HASH_SIZE))
		goto net_err;
	/* auth successful */
	alarm(0);
	PARA_INFO_LOG("good auth for %s\n", cc->u->name);
	/* init stream cipher keys with the second part of the random buffer */
	cc->scc.recv = sc_new(rand_buf + CHALLENGE_SIZE, SESSION_KEY_LEN,
		cf.aes_ctr128_requested);
	cc->scc.send = sc_new(rand_buf + CHALLENGE_SIZE + SESSION_KEY_LEN,
		SESSION_KEY_LEN, cf.aes_ctr128_requested);
	ret = send_sb(&cc->scc, NULL, 0, SBD_PROCEED, false);
	if (ret < 0)
		goto net_err;
	ret = recv_sb(&cc->scc, SBD_COMMAND, MAX_COMMAND_LEN, &iov);
	if (ret < 0)
		goto net_err;
	ret = run_command(cc, &iov, peername);
	free(iov.iov_base);
	if (ret < 0)
		goto err_out;
	if (ret >= 0)
		goto out;
err_out:
	if (send_strerror(cc, -ret) >= 0)
		send_sb(&cc->scc, NULL, 0, SBD_EXIT__FAILURE, true);
net_err:
	PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
out:
	free(buf);
	free(command);
	mutex_lock(mmd_mutex);
	mmd->active_connections--;
	mutex_unlock(mmd_mutex);
	if (ret >= 0) {
		ret = send_sb(&cc->scc, NULL, 0, SBD_EXIT__SUCCESS, true);
		if (ret < 0)
			PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
	}
	sc_free(cc->scc.recv);
	sc_free(cc->scc.send);
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}
